package core

import (
	"git.urbach.dev/cli/q/src/asm"
	"git.urbach.dev/cli/q/src/cpu"
	"git.urbach.dev/cli/q/src/errors"
	"git.urbach.dev/cli/q/src/expression"
	"git.urbach.dev/cli/q/src/token"
	"git.urbach.dev/cli/q/src/types"
)

// ExpressionToRegister puts the result of an expression into the specified register.
func (f *Function) ExpressionToRegister(node *expression.Expression, register cpu.Register) (types.Type, error) {
	if node.IsFolded {
		f.RegisterNumber(asm.MOVE, register, node.Value)
		return types.AnyInt, nil
	}

	if node.IsLeaf() {
		value, err := f.EvaluateToken(node.Token)

		if err != nil {
			return nil, err
		}

		f.ValueToRegister(value, register)
		return value.Type(), nil
	}

	switch node.Token.Kind {
	case token.Call:
		value, err := f.EvaluateCall(node)

		if err != nil {
			return nil, err
		}

		f.ValueToRegister(value, register)
		return value.Type(), nil

	case token.Dot:
		value, err := f.EvaluateDot(node)

		if err != nil {
			return nil, err
		}

		f.ValueToRegister(value, register)
		return value.Type(), nil

	case token.Array:
		value, err := f.EvaluateArray(node)

		if err != nil {
			return nil, err
		}

		f.ValueToRegister(value, register)
		return value.Type(), nil
	}

	if len(node.Children) == 1 {
		if !node.Token.IsUnaryOperator() {
			return nil, errors.New(errors.MissingOperand, f.File, node.Token.End())
		}

		typ, err := f.ExpressionToRegister(node.Children[0], register)

		if err != nil {
			return typ, err
		}

		return typ, f.ExecuteRegister(node.Token, register)
	}

	left := node.Children[0]
	right := node.Children[1]
	final := register

	if f.UsesRegister(right, register) {
		register = f.NewRegister()
	}

	typ, err := f.ExpressionToRegister(left, register)

	if err != nil {
		return nil, err
	}

	err = f.Execute(node.Token, register, right)

	if register != final {
		f.RegisterRegister(asm.MOVE, final, register)
		f.FreeRegister(register)
	}

	arrayType, isArray := typ.(*types.Array)

	if isArray {
		typ = &types.Pointer{To: arrayType.Of}
	} else if right.Token.Kind == token.Identifier {
		rightVariable := f.VariableByName(right.Token.Text(f.File.Bytes))
		leftPointer, leftIsPointer := typ.(*types.Pointer)
		rightPointer, rightIsPointer := rightVariable.Value.Typ.(*types.Pointer)

		if leftIsPointer && rightIsPointer && leftPointer.To == rightPointer.To {
			typ = types.Int
		}
	}

	return typ, err
}
